@@ -19,56 +19,65 @@
   describe('Launch Instance Model', function() {
 
     describe('launchInstanceModel Factory', function() {
-      var model, scope, settings, $q;
+      var model, scope, settings, $q, glance, IMAGE, VOLUME, VOLUME_SNAPSHOT, INSTANCE_SNAPSHOT;
       var cinderEnabled = false;
       var neutronEnabled = false;
       var novaExtensionsEnabled = false;
+
       var novaApi = {
         createServer: function(finalSpec) {
-            return {
-              then: function () {
-                return finalSpec;
-              }
-            };
-          },
+          return {
+            then: function () {
+              return finalSpec;
+            }
+          };
+        },
         getAvailabilityZones: function() {
-            var zones = [
-              { zoneName: 'zone-1', zoneState: { available: true } },
-              { zoneName: 'zone-2', zoneState: { available: true } },
-              { zoneName: 'invalid-zone-1' },
-              { zoneName: 'invalid-zone-2' }
-            ];
+          var zones = [
+            { zoneName: 'zone-1', zoneState: { available: true } },
+            { zoneName: 'zone-2', zoneState: { available: true } },
+            { zoneName: 'invalid-zone-1' },
+            { zoneName: 'invalid-zone-2' }
+          ];
 
-            var deferred = $q.defer();
-            deferred.resolve({ data: { items: zones } });
+          var deferred = $q.defer();
+          deferred.resolve({ data: { items: zones } });
 
-            return deferred.promise;
-          },
+          return deferred.promise;
+        },
         getFlavors: function() {
-            var flavors = [ 'flavor-1', 'flavor-2' ];
+          var flavors = [ 'flavor-1', 'flavor-2' ];
 
-            var deferred = $q.defer();
-            deferred.resolve({ data: { items: flavors } });
+          var deferred = $q.defer();
+          deferred.resolve({ data: { items: flavors } });
 
-            return deferred.promise;
-          },
+          return deferred.promise;
+        },
         getKeypairs: function() {
-            var keypairs = [ { keypair: { name: 'key-1' } },
-                             { keypair: { name: 'key-2' } } ];
+          var keypairs = [ { keypair: { name: 'key-1' } },
+                           { keypair: { name: 'key-2' } } ];
 
-            var deferred = $q.defer();
-            deferred.resolve({ data: { items: keypairs } });
+          var deferred = $q.defer();
+          deferred.resolve({ data: { items: keypairs } });
 
-            return deferred.promise;
-          },
+          return deferred.promise;
+        },
         getLimits: function() {
-            var limits = { maxTotalInstances: 10, totalInstancesUsed: 0 };
+          var limits = { maxTotalInstances: 10, totalInstancesUsed: 0 };
 
-            var deferred = $q.defer();
-            deferred.resolve({ data: limits });
+          var deferred = $q.defer();
+          deferred.resolve({ data: limits });
 
-            return deferred.promise;
-          }
+          return deferred.promise;
+        },
+        getServerGroups: function() {
+          var serverGroups = [ {'id': 'group-1'}, {'id': 'group-2'} ];
+
+          var deferred = $q.defer();
+          deferred.resolve({ data: { items: serverGroups } });
+
+          return deferred.promise;
+        }
       };
 
       var securityGroupApi = {
@@ -86,7 +95,11 @@
 
       var neutronApi = {
         getNetworks: function() {
-          var networks = [ { id: 'net-1' }, { id: 'net-2' } ];
+          var networks = [
+            { id: 'net-1', subnets: [ { id: 'subnet1' } ] },
+            { id: 'net-2', subnets: [ { id: 'subnet2' } ] },
+            { id: 'net-3', subnets: []}
+          ];
 
           var deferred = $q.defer();
           deferred.resolve({ data: { items: networks } });
@@ -102,7 +115,8 @@
             'net-2': [
               { name: 'port-3', device_owner: 'owner', fixed_ips: [], admin_state: 'DOWN' },
               { name: 'port-4', device_owner: '', fixed_ips: [], admin_state: 'DOWN' }
-            ]
+            ],
+            'net-3': []
           };
 
           var deferred = $q.defer();
@@ -114,6 +128,54 @@
 
       beforeEach(module('horizon.dashboard.project.workflow.launch-instance'));
 
+      beforeEach(module(function($provide) {
+        $provide.value('horizon.app.core.openstack-service-api.glance', {
+          getImages: function () {
+            var images = [
+              {container_format: 'aki', properties: {}},
+              {container_format: 'ari', properties: {}},
+              {container_format: 'ami', properties: {}},
+              {container_format: 'raw', properties: {}},
+              {container_format: 'ami', properties: {image_type: 'snapshot'}},
+              {container_format: 'raw', properties: {image_type: 'snapshot'}}
+            ];
+
+            var deferred = $q.defer();
+            deferred.resolve({data: {items: images}});
+
+            return deferred.promise;
+          },
+          getNamespaces: function () {
+            var namespaces = ['ns-1', 'ns-2'];
+
+            var deferred = $q.defer();
+            deferred.resolve({data: {items: namespaces}});
+
+            return deferred.promise;
+          }
+        });
+
+        beforeEach(function () {
+          settings = {
+            LAUNCH_INSTANCE_DEFAULTS: {
+              config_drive: false,
+              disable_image: false,
+              disable_instance_snapshot: false,
+              disable_volume: false,
+              disable_volume_snapshot: false
+            }
+          };
+          IMAGE = {type: 'image', label: 'Image'};
+          VOLUME = {type: 'volume', label: 'Volume'};
+          VOLUME_SNAPSHOT = {type: 'volume_snapshot', label: 'Volume Snapshot'};
+          INSTANCE_SNAPSHOT = {type: 'snapshot', label: 'Instance Snapshot'};
+        });
+
+        $provide.value('horizon.app.core.openstack-service-api.nova', novaApi);
+      }));
+
+      beforeEach(module('horizon.dashboard.project.workflow.launch-instance'));
+
       beforeEach(module(function($provide) {
         $provide.value('horizon.app.core.openstack-service-api.glance', {
           getImages: function() {
@@ -190,6 +252,23 @@
           }
         });
 
+        $provide.value('horizon.app.core.openstack-service-api.policy', {
+          ifAllowed: function() {
+            var deferred = $q.defer();
+
+            deferred.resolve();
+
+            return deferred.promise;
+          },
+          check: function() {
+            var deferred = $q.defer();
+
+            deferred.resolve();
+
+            return deferred.promise;
+          }
+        });
+
         $provide.value('horizon.app.core.openstack-service-api.novaExtensions', {
           ifNameEnabled: function() {
             var deferred = $q.defer();
@@ -210,6 +289,27 @@
 
             deferred.resolve(settings[setting]);
 
+            return deferred.promise;
+          },
+          ifEnabled: function(setting) {
+            var deferred = $q.defer();
+
+            var keys = setting.split('.');
+            var index = 0;
+            var value = settings;
+            while (angular.isObject(value) && index < keys.length) {
+              value = value[keys[index]];
+              index++;
+            }
+
+            // NOTE: This does not work for the general case of ifEnabled, only for what
+            // we need it for at the moment (only explicit false rejects the promise).
+            if (value === false) {
+              deferred.reject();
+            } else {
+              deferred.resolve();
+            }
+
             return deferred.promise;
           }
         });
@@ -219,10 +319,12 @@
         });
       }));
 
-      beforeEach(inject(function(launchInstanceModel, $rootScope, _$q_) {
-        model = launchInstanceModel;
-        $q = _$q_;
-        scope = $rootScope.$new();
+      beforeEach(inject(function($injector) {
+        model = $injector.get('launchInstanceModel');
+        $q = $injector.get('$q');
+        scope = $injector.get('$rootScope').$new();
+        glance = $injector.get('horizon.app.core.openstack-service-api.glance');
+        spyOn(glance, 'getNamespaces').and.callThrough();
       }));
 
       describe('Initial object (pre-initialize)', function() {
@@ -243,7 +345,8 @@
         it('has empty arrays for all data', function() {
           var datasets = ['availabilityZones', 'flavors', 'allowedBootSources',
             'images', 'imageSnapshots', 'keypairs', 'networks',
-            'profiles', 'securityGroups', 'volumes', 'volumeSnapshots'];
+            'profiles', 'securityGroups', 'serverGroups', 'volumes',
+            'volumeSnapshots'];
 
           datasets.forEach(function(name) {
             expect(model[name]).toEqual([]);
@@ -255,7 +358,8 @@
           expect(model.metadataDefs.image).toBeNull();
           expect(model.metadataDefs.volume).toBeNull();
           expect(model.metadataDefs.instance).toBeNull();
-          expect(Object.keys(model.metadataDefs).length).toBe(4);
+          expect(model.metadataDefs.hints).toBeNull();
+          expect(Object.keys(model.metadataDefs).length).toBe(5);
         });
 
         it('defaults "allow create volume from image" to false', function() {
@@ -274,6 +378,10 @@
           expect(model.metadataTree).toBe(null);
         });
 
+        it('defaults "hintsTree" to null', function() {
+          expect(model.hintsTree).toBe(null);
+        });
+
         it('initializes "nova limits" to empty object', function() {
           expect(model.novaLimits).toEqual({});
         });
@@ -358,7 +466,7 @@
         });
 
         it('should default config_drive to false if setting not provided', function() {
-          delete settings.LAUNCH_INSTANCE_DEFAULTS;
+          delete settings.LAUNCH_INSTANCE_DEFAULTS.config_drive;
           model.initialize(true);
           scope.$apply();
 
@@ -391,6 +499,19 @@
           expect(model.ports.length).toBe(1);
         });
 
+        it('should make 5 requests for namespaces', function() {
+          model.initialize(true);
+          scope.$apply();
+          expect(glance.getNamespaces.calls.count()).toBe(5);
+        });
+
+        it('should not request scheduler hints if scheduler hints disabled', function() {
+          settings.LAUNCH_INSTANCE_DEFAULTS.enable_scheduler_hints = false;
+          model.initialize(true);
+          scope.$apply();
+          expect(glance.getNamespaces.calls.count()).toBe(4);
+        });
+
         it('should set a keypair by default if only one keypair is available', function () {
           var keypair = { keypair: { name: 'key-1' } };
           spyOn(novaApi, 'getKeypairs').and.callFake(function () {
@@ -401,7 +522,7 @@
           model.initialize(true);
           scope.$apply();
           expect(model.newInstanceSpec.key_pair.length).toBe(1);
-          expect(model.newInstanceSpec.key_pair).toEqual( [ keypair.keypair ] );
+          expect(model.newInstanceSpec.key_pair).toEqual([ keypair.keypair ]);
         });
 
         it('should set a security group by default if one named "default" is available',
@@ -420,7 +541,7 @@
         );
 
         it('should set a network by default if only one network is available', function () {
-          var networks = [ { id: 'net-1' } ];
+          var networks = [ { id: 'net-1', subnets: [ { id: 'subnet1' } ] } ];
           spyOn(neutronApi, 'getNetworks').and.callFake(function () {
             var deferred = $q.defer();
             deferred.resolve({ data: { items: networks } });
@@ -431,6 +552,188 @@
           expect(model.newInstanceSpec.networks.length).toBe(1);
           expect(model.newInstanceSpec.networks).toEqual(networks);
         });
+
+        it('should have the proper entries in allowedBootSources', function() {
+          model.initialize(true);
+          scope.$apply();
+
+          expect(model.allowedBootSources).toBeDefined();
+          expect(model.allowedBootSources.length).toBe(4);
+          expect(model.allowedBootSources).toContain(IMAGE);
+          expect(model.allowedBootSources).toContain(VOLUME);
+          expect(model.allowedBootSources).toContain(VOLUME_SNAPSHOT);
+          expect(model.allowedBootSources).toContain(INSTANCE_SNAPSHOT);
+        });
+
+        it('should have proper allowedBootSources if settings are missing', function() {
+          delete settings.LAUNCH_INSTANCE_DEFAULTS;
+          model.initialize(true);
+          scope.$apply();
+
+          expect(model.allowedBootSources).toBeDefined();
+          expect(model.allowedBootSources.length).toBe(4);
+          expect(model.allowedBootSources).toContain(IMAGE);
+          expect(model.allowedBootSources).toContain(INSTANCE_SNAPSHOT);
+          expect(model.allowedBootSources).toContain(VOLUME);
+          expect(model.allowedBootSources).toContain(VOLUME_SNAPSHOT);
+        });
+
+        it('should have proper allowedBootSources if specific settings missing', function() {
+          delete settings.LAUNCH_INSTANCE_DEFAULTS.disable_image;
+          delete settings.LAUNCH_INSTANCE_DEFAULTS.disable_instance_snapshot;
+          delete settings.LAUNCH_INSTANCE_DEFAULTS.disable_volume;
+          model.initialize(true);
+          scope.$apply();
+
+          expect(model.allowedBootSources).toBeDefined();
+          expect(model.allowedBootSources.length).toBe(4);
+          expect(model.allowedBootSources).toContain(IMAGE);
+          expect(model.allowedBootSources).toContain(INSTANCE_SNAPSHOT);
+          expect(model.allowedBootSources).toContain(VOLUME);
+          expect(model.allowedBootSources).toContain(VOLUME_SNAPSHOT);
+        });
+
+        it('should have no images if disable_image is set to true', function() {
+          settings.LAUNCH_INSTANCE_DEFAULTS.disable_image = true;
+          model.initialize(true);
+          scope.$apply();
+
+          expect(model.images.length).toBe(0);
+          expect(model.images).toEqual([]);
+          expect(model.allowedBootSources).toBeDefined();
+          expect(model.allowedBootSources.length).toBe(3);
+          expect(model.allowedBootSources).not.toContain(IMAGE);
+          expect(model.allowedBootSources).toContain(INSTANCE_SNAPSHOT);
+          expect(model.allowedBootSources).toContain(VOLUME);
+          expect(model.allowedBootSources).toContain(VOLUME_SNAPSHOT);
+        });
+
+        it('should have images if disable_image is missing', function() {
+          delete settings.LAUNCH_INSTANCE_DEFAULTS.disable_image;
+          model.initialize(true);
+          scope.$apply();
+
+          expect(model.allowedBootSources).toBeDefined();
+          expect(model.allowedBootSources.length).toBe(4);
+          expect(model.allowedBootSources).toContain(IMAGE);
+          expect(model.allowedBootSources).toContain(INSTANCE_SNAPSHOT);
+          expect(model.allowedBootSources).toContain(VOLUME);
+          expect(model.allowedBootSources).toContain(VOLUME_SNAPSHOT);
+        });
+
+        it('should have no volumes if disable_volume is set to true', function() {
+          settings.LAUNCH_INSTANCE_DEFAULTS.disable_volume = true;
+          model.initialize(true);
+          scope.$apply();
+
+          expect(model.volumes.length).toBe(0);
+          expect(model.volumes).toEqual([]);
+          expect(model.volumeSnapshots.length).toBe(2);
+          expect(model.volumeSnapshots).toEqual([{ id: 'snap-1' }, { id: 'snap-2' }]);
+          expect(model.allowedBootSources).toBeDefined();
+          expect(model.allowedBootSources.length).toBe(3);
+          expect(model.allowedBootSources).toContain(IMAGE);
+          expect(model.allowedBootSources).toContain(INSTANCE_SNAPSHOT);
+          expect(model.allowedBootSources).not.toContain(VOLUME);
+          expect(model.allowedBootSources).toContain(VOLUME_SNAPSHOT);
+        });
+
+        it('should have volumes if disable_volume is missing', function() {
+          delete settings.LAUNCH_INSTANCE_DEFAULTS.disable_volume;
+          model.initialize(true);
+          scope.$apply();
+
+          expect(model.allowedBootSources).toBeDefined();
+          expect(model.allowedBootSources.length).toBe(4);
+          expect(model.allowedBootSources).toContain(IMAGE);
+          expect(model.allowedBootSources).toContain(INSTANCE_SNAPSHOT);
+          expect(model.allowedBootSources).toContain(VOLUME);
+          expect(model.allowedBootSources).toContain(VOLUME_SNAPSHOT);
+        });
+
+        it('should have volume snapshots if disable_volume_snapshot is missing', function() {
+          delete settings.LAUNCH_INSTANCE_DEFAULTS.disable_volume_snapshot;
+          model.initialize(true);
+          scope.$apply();
+
+          expect(model.allowedBootSources).toBeDefined();
+          expect(model.allowedBootSources.length).toBe(4);
+          expect(model.allowedBootSources).toContain(IMAGE);
+          expect(model.allowedBootSources).toContain(INSTANCE_SNAPSHOT);
+          expect(model.allowedBootSources).toContain(VOLUME);
+          expect(model.allowedBootSources).toContain(VOLUME_SNAPSHOT);
+        });
+
+        it('should not have volume snapshots if disable_volume_snapshot is set to true',
+        function() {
+          settings.LAUNCH_INSTANCE_DEFAULTS.disable_volume_snapshot = true;
+          model.initialize(true);
+          scope.$apply();
+
+          expect(model.allowedBootSources).toBeDefined();
+          expect(model.allowedBootSources.length).toBe(3);
+          expect(model.allowedBootSources).toContain(IMAGE);
+          expect(model.allowedBootSources).toContain(INSTANCE_SNAPSHOT);
+          expect(model.allowedBootSources).toContain(VOLUME);
+          expect(model.allowedBootSources).not.toContain(VOLUME_SNAPSHOT);
+        });
+
+        it('should have no snapshot if disable_instance_snapshot is set to true', function() {
+          settings.LAUNCH_INSTANCE_DEFAULTS.disable_instance_snapshot = true;
+          model.initialize(true);
+          scope.$apply();
+
+          expect(model.allowedBootSources).toBeDefined();
+          expect(model.allowedBootSources.length).toBe(3);
+          expect(model.allowedBootSources).toContain(IMAGE);
+          expect(model.allowedBootSources).not.toContain(INSTANCE_SNAPSHOT);
+          expect(model.allowedBootSources).toContain(VOLUME);
+          expect(model.allowedBootSources).toContain(VOLUME_SNAPSHOT);
+        });
+
+        it('should have snapshot if disable_instance_snapshot is missing', function() {
+          delete settings.LAUNCH_INSTANCE_DEFAULTS.disable_instance_snapshot;
+          model.initialize(true);
+          scope.$apply();
+
+          expect(model.allowedBootSources).toBeDefined();
+          expect(model.allowedBootSources.length).toBe(4);
+          expect(model.allowedBootSources).toContain(IMAGE);
+          expect(model.allowedBootSources).toContain(INSTANCE_SNAPSHOT);
+          expect(model.allowedBootSources).toContain(VOLUME);
+          expect(model.allowedBootSources).toContain(VOLUME_SNAPSHOT);
+        });
+
+        it('should have no snapshot and no image if both are disabled', function() {
+          settings.LAUNCH_INSTANCE_DEFAULTS.disable_image = true;
+          settings.LAUNCH_INSTANCE_DEFAULTS.disable_instance_snapshot = true;
+
+          model.initialize(true);
+          scope.$apply();
+
+          expect(model.allowedBootSources).toBeDefined();
+          expect(model.allowedBootSources.length).toBe(2);
+          expect(model.allowedBootSources).not.toContain(IMAGE);
+          expect(model.allowedBootSources).not.toContain(INSTANCE_SNAPSHOT);
+          expect(model.allowedBootSources).toContain(VOLUME);
+          expect(model.allowedBootSources).toContain(VOLUME_SNAPSHOT);
+        });
+
+        it('should have snapshot and image if both are missing', function() {
+          delete settings.LAUNCH_INSTANCE_DEFAULTS.disable_image;
+          delete settings.LAUNCH_INSTANCE_DEFAULTS.disable_instance_snapshot;
+
+          model.initialize(true);
+          scope.$apply();
+
+          expect(model.allowedBootSources).toBeDefined();
+          expect(model.allowedBootSources.length).toBe(4);
+          expect(model.allowedBootSources).toContain(IMAGE);
+          expect(model.allowedBootSources).toContain(INSTANCE_SNAPSHOT);
+          expect(model.allowedBootSources).toContain(VOLUME);
+          expect(model.allowedBootSources).toContain(VOLUME_SNAPSHOT);
+        });
+
       });
 
       describe('Post Initialization Model - Initializing', function() {
@@ -443,7 +746,7 @@
         // This is here to ensure that as people add/change items, they
         // don't forget to implement tests for them.
         it('has the right number of properties', function() {
-          expect(Object.keys(model.newInstanceSpec).length).toBe(19);
+          expect(Object.keys(model.newInstanceSpec).length).toBe(21);
         });
 
         it('sets availability zone to null', function() {
@@ -498,6 +801,10 @@
           expect(model.newInstanceSpec.security_groups).toEqual([]);
         });
 
+        it('sets scheduler hints to an empty object', function() {
+          expect(model.newInstanceSpec.scheduler_hints).toEqual({});
+        });
+
         it('sets source type to null', function() {
           expect(model.newInstanceSpec.source_type).toBeNull();
         });
@@ -516,7 +823,7 @@
       });
 
       describe('Create Instance', function() {
-        var metadata;
+        var metadata, hints;
 
         beforeEach(function() {
           // initialize some data
@@ -528,10 +835,12 @@
           model.newInstanceSpec.key_pair = [ { name: 'keypair1' } ];
           model.newInstanceSpec.security_groups = [ { id: 'adminId', name: 'admin' },
                                                     { id: 'demoId', name: 'demo' } ];
+          model.newInstanceSpec.scheduler_hints = {};
           model.newInstanceSpec.vol_create = true;
           model.newInstanceSpec.vol_delete_on_instance_delete = true;
           model.newInstanceSpec.vol_device_name = "volTestName";
           model.newInstanceSpec.vol_size = 10;
+          model.newInstanceSpec.server_groups = [];
 
           metadata = {'foo': 'bar'};
           model.metadataTree = {
@@ -539,6 +848,13 @@
               return metadata;
             }
           };
+
+          hints = {'hint1': 'val1'};
+          model.hintsTree = {
+            getExisting: function() {
+              return hints;
+            }
+          };
         });
 
         it('should set final spec in format required by Nova (Neutron disabled)', function() {
@@ -693,6 +1009,36 @@
           expect(finalSpec.meta).toBe(metadata);
         });
 
+        it('should have only group for scheduler_hints if no other hints specified', function() {
+          hints = {};
+          model.newInstanceSpec.server_groups = [{'id': 'group1'}];
+          var finalHints = {'group': model.newInstanceSpec.server_groups[0].id};
+
+          var finalSpec = model.createInstance();
+          expect(finalSpec.scheduler_hints).toEqual(finalHints);
+
+          model.hintsTree = null;
+
+          finalSpec = model.createInstance();
+          expect(finalSpec.scheduler_hints).toEqual(finalHints);
+        });
+
+        it('should have scheduler_hints property if scheduler hints specified', function() {
+          var finalHints = hints;
+          finalHints.group = 'group1';
+
+          var finalSpec = model.createInstance();
+          expect(finalSpec.scheduler_hints).toEqual(finalHints);
+        });
+
+        it('should have no scheduler_hints if no scheduler hints specified', function() {
+          hints = {};
+          model.newInstanceSpec.server_groups = [];
+
+          var finalSpec = model.createInstance();
+          expect(finalSpec.scheduler_hints).toEqual({});
+        });
+
       });
     });
   });
